"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.times import Duration
from encommon.types import DictStrAny
from encommon.types import sort_dict
from encommon.types.strings import NEWLINE
from encommon.types.strings import SEMPTY

if TYPE_CHECKING:
    from ...robie.childs import RobieClient
    from ...robie.models import RobieMessage
    from .common import StatusPluginItem
    from .params import StatusPluginReportParams
    from .plugin import StatusPlugin



def grouped(
    plugin: 'StatusPlugin',
) -> dict[str, list['StatusPluginItem']]:
    """
    Return the dictionary with status value stored by group.

    :param plugin: Plugin class instance for Chatting Robie.
    :returns: Dictionary with status value stored by group.
    """

    status = plugin.status

    groups: DictStrAny = {
        x.group: [] for x
        in status.values()}

    items = status.items()

    for name, value in items:

        group = value.group

        (groups[group]
         .append(value))

    return sort_dict(groups)



def composedsc(
    plugin: 'StatusPlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue

    compose: list[str] = []


    def _compose(
        message: str,
    ) -> None:

        citem = mitem.reply(
            robie, message)

        cqueue.put(citem)


    def _format() -> None:

        epoch = int(value.time)

        sicon = (
            (getattr(
                params.icons,
                value.state)
             .dsc)
            or SEMPTY)

        sicon = (
            f'{sicon} '
            if len(sicon) >= 1
            else sicon)

        vicon = (
            value.icon.dsc
            or SEMPTY)

        compose.append(
            (f'{vicon or ""} '
             f'{value.title}/'
             f'`{value.unique}`:'
             f' {sicon or ""}'
             f'`{value.state}` '
             f'<t:{epoch}:R>')
            .strip())


    items = sorted(
        grouped(plugin)
        .items())

    for name, values in items:

        compose.append(
            f'**{name}**')

        _values = sorted(values)

        for value in _values:
            _format()

    _compose(
        NEWLINE
        .join(compose))



def reportdsc(
    plugin: 'StatusPlugin',
    client: 'RobieClient',
    status: 'StatusPluginItem',
    report: 'StatusPluginReportParams',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param client: Client class instance for Chatting Robie.
    :param status: Object containing the status information.
    :param report: Object containing the report information.
    """

    assert plugin.thread

    thread = plugin.thread
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue


    def _compose(
        message: str,
    ) -> None:

        citem = (
            client.compose(
                report.target,
                message))

        cqueue.put(citem)


    epoch = int(status.time)

    sicon = (
        (getattr(
            params.icons,
            status.state)
         .dsc)
        or SEMPTY)

    sicon = (
        f'{sicon} '
        if len(sicon) >= 1
        else sicon)

    vicon = (
        status.icon.dsc
        or SEMPTY)

    _compose(
        (f'{vicon or ""} '
         f'{status.title}/'
         f'`{status.unique}`:'
         f' {sicon or ""}'
         f'`{status.state}` '
         f'<t:{epoch}:R>')
        .strip())



def composeirc(
    plugin: 'StatusPlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue


    def _compose(
        message: str,
    ) -> None:

        citem = mitem.reply(
            robie, message)

        cqueue.put(citem)


    def _format() -> None:

        durate = Duration(
            value.time.since)

        sicon = (
            (getattr(
                params.icons,
                value.state)
             .irc)
            or SEMPTY)

        sicon = (
            f'{sicon} '
            if len(sicon) >= 1
            else sicon)

        vicon = (
            value.icon.irc
            or SEMPTY)

        _compose(
            (f'{vicon or ""} '
             f'{value.title}/'
             f'{value.unique}:'
             f' {sicon or ""}'
             f'{value.state} '
             f'for {durate}')
            .strip())


    items = sorted(
        grouped(plugin)
        .items())

    for name, values in items:

        _compose(name)

        _values = sorted(values)

        for value in _values:
            _format()



def reportirc(
    plugin: 'StatusPlugin',
    client: 'RobieClient',
    status: 'StatusPluginItem',
    report: 'StatusPluginReportParams',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param client: Client class instance for Chatting Robie.
    :param status: Object containing the status information.
    :param report: Object containing the report information.
    """

    assert plugin.thread

    thread = plugin.thread
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue


    def _compose(
        message: str,
    ) -> None:

        citem = (
            client.compose(
                report.target,
                message))

        cqueue.put(citem)


    durate = Duration(
        status.time.since)

    sicon = (
        (getattr(
            params.icons,
            status.state)
         .irc)
        or SEMPTY)

    sicon = (
        f'{sicon} '
        if len(sicon) >= 1
        else sicon)

    vicon = (
        status.icon.irc
        or SEMPTY)

    _compose(
        (f'{vicon or ""} '
         f'{status.title}/'
         f'{status.unique}:'
         f' {sicon or ""}'
         f'{status.state} '
         f'for {durate}')
        .strip())



def composemtm(
    plugin: 'StatusPlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue

    compose: list[str] = []


    def _compose(
        message: str,
    ) -> None:

        citem = mitem.reply(
            robie, message)

        cqueue.put(citem)


    def _format() -> None:

        durate = Duration(
            value.time.since)

        sicon = (
            (getattr(
                params.icons,
                value.state)
             .mtm)
            or SEMPTY)

        sicon = (
            f'{sicon} '
            if len(sicon) >= 1
            else sicon)

        vicon = (
            value.icon.mtm
            or SEMPTY)

        compose.append(
            (f'{vicon or ""} '
             f'{value.title}/'
             f'`{value.unique}`:'
             f' {sicon or ""}'
             f'`{value.state}` '
             f'for `{durate}`')
            .strip())


    items = sorted(
        grouped(plugin)
        .items())

    for name, values in items:

        compose.append(
            f'**{name}**')

        _values = sorted(values)

        for value in _values:
            _format()

    _compose(
        NEWLINE
        .join(compose))



def reportmtm(
    plugin: 'StatusPlugin',
    client: 'RobieClient',
    status: 'StatusPluginItem',
    report: 'StatusPluginReportParams',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param client: Client class instance for Chatting Robie.
    :param status: Object containing the status information.
    :param report: Object containing the report information.
    """

    assert plugin.thread

    thread = plugin.thread
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue


    def _compose(
        message: str,
    ) -> None:

        citem = (
            client.compose(
                report.target,
                message))

        cqueue.put(citem)


    durate = Duration(
        status.time.since)

    sicon = (
        (getattr(
            params.icons,
            status.state)
         .mtm)
        or SEMPTY)

    sicon = (
        f'{sicon} '
        if len(sicon) >= 1
        else sicon)

    vicon = (
        status.icon.mtm
        or SEMPTY)

    _compose(
        (f'{vicon or ""} '
         f'{status.title}/'
         f'`{status.unique}`:'
         f' {sicon or ""}'
         f'`{status.state}` '
         f'for `{durate}`')
        .strip())



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..importer import importer



def test_importer() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    dumps = importer('json.dumps')

    assert dumps([1]) == '[1]'



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated

from encommon.types import BaseModel

from pydantic import Field



class AinswerResponse(BaseModel, extra='forbid'):
    """
    Contains the response for question for the client type.
    """

    text: Annotated[
        str,
        Field(...,
              description='Simple and concise format.')]



class AinswerResponseDSC(AinswerResponse, extra='forbid'):
    """
    Contains the response for question for the client type.
    """

    text: Annotated[
        str,
        Field(...,
              description='Format intended for Discord.',
              max_length=1900)]



class AinswerResponseIRC(AinswerResponse, extra='forbid'):
    """
    Contains the response for question for the client type.
    """

    text: Annotated[
        str,
        Field(...,
              description='Format intended for IRCv2.',
              max_length=350)]



class AinswerResponseMTM(AinswerResponse, extra='forbid'):
    """
    Contains the response for question for the client type.
    """

    text: Annotated[
        str,
        Field(...,
              description='Format intended for Mattermost.',
              max_length=1900)]



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional

from pydantic import Field

from .child import RobieChildParams



class RobieClientParams(RobieChildParams, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    locate: Annotated[
        Optional[str],
        Field(None,
              description='For which client are parameters',
              examples=[
                  'enrobie.clients.DSCClient',
                  'enrobie.clients.IRCClient',
                  'enrobie.clients.MTMClient'],
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional

from encommon.config import Params

from pydantic import Field

from .client import RobieClientParams
from .common import RobieParamsModel
from .person import RobiePersonParams
from .plugin import RobiePluginParams
from .service import RobieServiceParams



class RobiePrinterParams(RobieParamsModel, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    message: Annotated[
        bool,
        Field(False,
              description='Print the messages to console')]

    command: Annotated[
        bool,
        Field(False,
              description='Print the commands to console')]



class RobieParams(Params, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    database: Annotated[
        str,
        Field('sqlite:///:memory:',
              description='Database connection string',
              min_length=1)]

    printer: Annotated[
        RobiePrinterParams,
        Field(default_factory=RobiePrinterParams,
              description='Print messages to console')]

    service: Annotated[
        RobieServiceParams,
        Field(default_factory=RobieServiceParams,
              description='Parameters for Robie Service')]

    clients: Annotated[
        Optional[dict[str, RobieClientParams]],
        Field(None,
              description='Parameters for Robie clients',
              min_length=1)]

    plugins: Annotated[
        Optional[dict[str, RobiePluginParams]],
        Field(None,
              description='Parameters for Robie plugins',
              min_length=1)]

    persons: Annotated[
        Optional[dict[str, RobiePersonParams]],
        Field(None,
              description='Parameters for Robie persons',
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        if _parse is not None:

            parsable = ['persons']

            for key in parsable:

                if not data.get(key):
                    continue

                values = (
                    data[key]
                    .values())

                for item in values:
                    item['_parse'] = _parse


        if _parse is not None:

            parsable = [
                'database',
                'printer',
                'service']

            for key in parsable:

                value = data.get(key)

                if value is None:
                    continue

                value = _parse(value)

                data[key] = value


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.types import DictStrAny
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils.sample import ENPYRWS

from . import SAMPLES
from ..config import RobieConfig
from ..models import RobieMessage
from ..robie import Robie



def test_Robie(
    robie: Robie,
    replaces: DictStrAny,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    :param replaces: Mapping of what to replace in samples.
    """


    attrs = lattrs(robie)

    assert attrs == [
        '_Robie__config',
        '_Robie__logger',
        '_Robie__jinja2',
        '_Robie__childs']


    assert inrepr(
        'robie.Robie',
        robie)

    assert isinstance(
        hash(robie), int)

    assert instr(
        'robie.Robie',
        robie)


    assert robie.config

    assert robie.logger

    assert robie.jinja2

    assert robie.childs

    assert robie.params


    sample_path = (
        SAMPLES / 'dumped.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=robie.dumped,
        replace=replaces)

    expect = prep_sample(
        content=robie.dumped,
        replace=replaces)

    assert expect == sample



def test_Robie_printer(
    robie: Robie,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']

    item = RobieMessage(client)

    robie.printer(item)



def test_Robie_person(
    robie: Robie,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients


    person = robie.person(
        clients['dscbot'],
        '823902304920392013')

    assert person is not None
    assert person.name == 'bender'

    person = robie.person(
        clients['dscbot'], '123')

    assert person is None


    person = robie.person(
        clients['ircbot'],
        'bender!bender@bending.com')

    assert person is not None
    assert person.name == 'bender'

    person = robie.person(
        clients['ircbot'],
        'some!user@domain.invalid')

    assert person is None


    person = robie.person(
        clients['mtmbot'],
        'iietalkwkjf9al2kla')

    assert person is not None
    assert person.name == 'bender'

    person = robie.person(
        clients['mtmbot'], '123')

    assert person is None



def test_Robie_jinja2(
    robie: Robie,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    parsed = robie.j2parse(
        '{{ foo }}',
        {'foo': 'bar'})

    assert parsed == 'bar'



def test_Robie_cover(
) -> None:
    """
    Perform various tests associated with relevant routines.
    """

    Robie(RobieConfig())



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils.sample import ENPYRWS

from enconnect.discord import ClientEvent
from enconnect.fixtures import DSCClientSocket

from pytest import raises

from . import SAMPLES
from .test_client import DSCEVENTS
from .test_client import DSCEVENT_HUBERT_CHAN
from .test_client import DSCEVENT_RANDOM_CHAN
from .test_client import DSCEVENT_RANDOM_PRIV
from ..client import DSCClient
from ..command import DSCCommand
from ..message import DSCMessage

if TYPE_CHECKING:
    from ....robie import Robie
    from ....robie import RobieService
    from ....robie.models import RobieMessage



def test_DSCMessage(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['dscbot']

    assert isinstance(
        client, DSCClient)


    item = DSCMessage(
        client,
        ClientEvent(
            client.client,
            DSCEVENT_HUBERT_CHAN))


    attrs = lattrs(item)

    assert attrs == [
        'event',
        'client',
        'person',
        'time']


    assert inrepr(
        'DSCMessage',
        item)

    with raises(TypeError):
        hash(item)

    assert instr(
        'DSCMessage',
        item)


    assert item.time.since > 0

    assert item.client == client.name

    assert item.family == 'discord'

    assert item.kind == 'chanmsg'

    assert item.person == 'hubert'

    assert not item.isme

    assert not item.hasme

    assert not item.whome

    assert item.author
    assert item.author[0] == 'hubert'

    assert item.anchor == 'guildid:enrobie'

    assert item.message


    event = item.event

    assert event.type == (
        'MESSAGE_CREATE')
    assert event.opcode == 0
    assert event.data
    assert len(event.data) == 4
    assert event.seqno == 7
    assert len(event.original) == 4

    assert event.kind == 'chanmsg'
    assert not event.isme
    assert not event.hasme
    assert not event.whome
    assert event.author == (
        'hubert', '823039201390230492')
    assert event.recipient == (
        'guildid', 'enrobie')
    assert event.message == 'dscbot'



def test_DSCMessage_reply(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['dscbot']

    assert isinstance(
        client, DSCClient)


    item = DSCMessage(
        client,
        ClientEvent(
            client.client,
            DSCEVENT_RANDOM_CHAN))

    reply = item.reply(
        robie, 'Hello')

    assert isinstance(
        reply, DSCCommand)

    assert reply.json == {
        'content': 'Hello'}


    item = DSCMessage(
        client,
        ClientEvent(
            client.client,
            DSCEVENT_RANDOM_PRIV))

    reply = item.reply(
        robie, 'Hello')

    assert isinstance(
        reply, DSCCommand)

    assert reply.json == {
        'content': 'Hello'}



def test_DSCMessage_samples(
    tmp_path: Path,
    service: 'RobieService',
    client_dscsock: DSCClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param service: Ancilary Chatting Robie class instance.
    :param client_dscsock: Object to mock client connection.
    """

    robie = service.robie
    childs = robie.childs
    clients = childs.clients

    client = clients['dscbot']

    assert isinstance(
        client, DSCClient)


    messages: list[DSCMessage] = []


    def _callback(
        mitem: 'RobieMessage',
    ) -> None:

        assert isinstance(
            mitem, DSCMessage)

        messages.append(mitem)


    (client.publish
     .subscribe(_callback))


    client_dscsock(DSCEVENTS)

    service.limit(
        clients=['dscbot'],
        plugins=['status'])

    service.start()


    thread = Thread(
        target=service.operate)

    thread.start()


    block_sleep(5)


    sample_path = (
        SAMPLES / 'messages.json')

    # Because of ones in helper
    maximum = len(DSCEVENTS) + 4

    _messages = [
        {'client': x.client,
         'family': x.family,
         'kind': x.kind,
         'person': x.person,
         'isme': x.isme,
         'hasme': x.hasme,
         'whome': x.whome,
         'author': x.author,
         'anchor': x.anchor,
         'message': x.message,
         'event': x.event}
        for x in
        messages[:maximum]]

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=_messages)

    expect = prep_sample(
        content=_messages)

    assert expect == sample


    service.soft()

    while service.running:
        block_sleep(1)

    service.stop()

    thread.join()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..plugin import LoggerPlugin
from ....clients import IRCClient

if TYPE_CHECKING:
    from ....robie import Robie



def _logger_history(
    plugin: LoggerPlugin,
    client: 'IRCClient',
) -> None:
    """
    Insert testing records into the provided history object.

    :param plugin: Plugin class instance for Chatting Robie.
    :param client: Client class instance for Chatting Robie.
    """

    history = plugin.history


    for count in range(4):

        count += 1

        for nick in range(4):

            nick += 1

            history.insert(
                client=client.name,
                person=None,
                kind='chanmsg',
                author=f'nickname{nick}',
                anchor='#enrobie',
                message=f'Message {count}')

            history.insert(
                client=client.name,
                person=None,
                kind='privmsg',
                author=f'nickname{nick}',
                anchor=f'nickname{nick}',
                message=f'Message {count}')

            block_sleep(0.001)


    history.insert(
        client=client.name,
        person='hubert',
        kind='chanmsg',
        author='hubert',
        anchor='#enrobie',
        message='Good news')

    history.insert(
        client=client.name,
        person='hubert',
        kind='privmsg',
        author='hubert',
        anchor='hubert',
        message='Good news')



def test_LoggerHistory(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    plugins = childs.plugins

    plugin = plugins['logger']

    assert isinstance(
        plugin, LoggerPlugin)


    history = plugin.history


    attrs = lattrs(history)

    assert attrs == [
        '_LoggerHistory__plugin',
        '_LoggerHistory__connect',
        '_LoggerHistory__locker',
        '_LoggerHistory__sengine',
        '_LoggerHistory__session']


    assert inrepr(
        'history.LoggerHistory',
        history)

    assert isinstance(
        hash(history), int)

    assert instr(
        'history.LoggerHistory',
        history)



def test_LoggerHistory_cover(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients
    plugins = childs.plugins

    client = clients['ircbot']
    plugin = plugins['logger']

    assert isinstance(
        client, IRCClient)

    assert isinstance(
        plugin, LoggerPlugin)

    history = plugin.history


    _logger_history(
        plugin, client)


    records = (
        history.search(
            client=client.name,
            anchor='#enrobie'))

    assert len(records) == 10


    record = (
        records[0].endumped)

    assert record == {
        'anchor': '#enrobie',
        'author': 'nickname4',
        'client': 'ircbot',
        'create': record['create'],
        'kind': 'chanmsg',
        'message': 'Message 2',
        'person': None,
        'plugin': 'logger'}


    records = (
        history.search(
            client=client.name,
            anchor='nickname1'))

    record = (
        records[-1].endumped)

    assert record == {
        'anchor': 'nickname1',
        'author': 'nickname1',
        'client': 'ircbot',
        'create': record['create'],
        'kind': 'privmsg',
        'message': 'Message 4',
        'person': None,
        'plugin': 'logger'}


    records = (
        history.search(
            client=client.name,
            author='nickname1',
            anchor='nickname1',
            limit=1))

    assert len(records) == 1


    plaintext = (
        history.plaintext(
            client=client.name,
            anchor='nickname1'))

    plain = plaintext[-1]

    assert plain[27:] == (
        '<nickname1> Message 4')



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from enconnect.fixtures import IRCClientSocket
from enconnect.irc import ClientEvent
from enconnect.irc.test import EVENTS

from ..client import IRCClient
from ..command import IRCCommand
from ....robie.addons import RobieQueue

if TYPE_CHECKING:
    from ....robie import Robie
    from ....robie import RobieService
    from ....robie.models import RobieCommand  # noqa: F401
    from ....robie.models import RobieMessage  # noqa: F401



IRCEVENT_RANDOM_CHAN = (
    ':nick!user@host PRIVMSG'
    ' #enrobie :Hello ircbot')

IRCEVENT_RANDOM_PRIV = (
    ':nick!user@host PRIVMSG'
    ' ircbot :Hello ircbot')

IRCEVENT_HUBERT_CHAN = (
    ':hubert!hubert@science.com'
    ' PRIVMSG #enrobie :ircbot')

IRCEVENT_HUBERT_PRIV = (
    ':hubert!hubert@science.com'
    ' PRIVMSG ircbot :ircbot')

IRCEVENTS: list[str] = [

    *EVENTS[:-1],

    (':botirc!user@host'
     ' NICK :ircbot'),

    ':ircbot JOIN :#enrobie',

    (':mocked 353 ircbot = #enrobie'
     ' :ircbot robert trebor sirrah'),

    (':mocked 332 ircbot #enrobie'
     ' :Test topic already set'),

    (':nick!user@host TOPIC #enrobie'
     ' :Test topic is changed'),

    (':nick!user@host KICK '
     '#enrobie ircbot :foo'),

    ':ircbot JOIN :#enrobie',

    (':mocked 353 ircbot = #enrobie'
     ' :ircbot robert trebor sirrah'),

    (':mocked 332 ircbot #enrobie'
     ' :Test topic is changed'),

    # From random to channel
    IRCEVENT_RANDOM_CHAN,

    # From random to private
    IRCEVENT_RANDOM_PRIV,

    # From hubert to channel
    IRCEVENT_HUBERT_CHAN,

    # From hubert to channel
    IRCEVENT_HUBERT_CHAN,

    # From hubert to private
    IRCEVENT_HUBERT_PRIV,

    (':nick!user@host TOPIC #enrobie'
     ' :Test topic'),

    (':trebor!user@host NICK treb0r'),

    (':nick!user@host NICK n1ck'),

    ':ircbot PART :#enrobie']



def test_IRCClient(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    client = (
        service.clients
        .childs['ircbot'])

    assert isinstance(
        client,
        IRCClient)


    attrs = lattrs(client)

    assert attrs == [
        '_RobieChild__robie',
        '_RobieChild__name',
        '_RobieChild__params',
        '_IRCClient__client',
        '_IRCClient__channels',
        '_IRCClient__publish',
        '_RobieClient__thread']


    assert inrepr(
        'client.IRCClient',
        client)

    assert isinstance(
        hash(client), int)

    assert instr(
        'client.IRCClient',
        client)


    client.validate()

    assert client.robie

    assert client.enable

    assert client.name == 'ircbot'

    assert client.family == 'irc'

    assert client.kind == 'client'

    assert client.client

    assert client.channels

    assert client.publish

    assert client.schema()

    assert client.params

    assert client.thread

    assert client.dumped



def test_IRCClient_message(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']

    assert isinstance(
        client, IRCClient)


    _queue = RobieQueue[
        'RobieMessage']

    queue: _queue = (
        RobieQueue(robie))


    _event = 'PING :123456789'

    event = ClientEvent(
        client.client, _event)

    client.put_message(
        queue, event)



def test_IRCClient_command(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']

    assert isinstance(
        client, IRCClient)


    _queue = RobieQueue[
        'RobieCommand']

    queue: _queue = (
        RobieQueue(robie))


    client.put_command(
        queue,
        'PRIVMSG # :Hello')



def test_IRCClient_compose(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']

    assert isinstance(
        client, IRCClient)


    citem = (
        client.compose(
            '#enrobie',
            'message'))

    assert isinstance(
        citem, IRCCommand)


    assert citem.event == (
        'PRIVMSG #enrobie'
        ' :message')



def test_IRCClient_channels(
    service: 'RobieService',
    client_ircsock: IRCClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    :param client_ircsock: Object to mock client connection.
    """

    robie = service.robie
    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']

    assert isinstance(
        client, IRCClient)


    client_ircsock(IRCEVENTS)


    service.limit(
        clients=['ircbot'],
        plugins=['status'])

    service.start()


    thread = Thread(
        target=service.operate)

    thread.start()


    block_sleep(5)


    select = (
        client.channels
        .select('#enrobie'))

    assert select is not None

    assert select.endumped == {
        'members': {
            'robert',
            'sirrah',
            'treb0r'},
        'title': '#enrobie',
        'topic': 'Test topic',
        'unique': '#enrobie'}


    service.soft()

    while service.running:
        block_sleep(1)

    service.stop()

    thread.join()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import Type
from typing import Union

from encommon.types import NCNone

from .current import NagiosCurrent
from .helpers import composedsc
from .helpers import composeirc
from .helpers import composemtm
from .params import NagiosPluginParams
from ..ainswer import AinswerTool
from ..status import StatusPlugin
from ..status import StatusPluginStates
from ...robie.childs import RobiePerson
from ...robie.childs import RobiePlugin
from ...robie.models import RobieMessage



class NagiosPlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin allows for connecting to Nagios Console.
    """

    __started: bool

    __current: NagiosCurrent


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__started = False

        self.__current = (
            NagiosCurrent(self))

        self.__status('pending')


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Review the parameters


    @classmethod
    def schema(
        cls,
    ) -> Type[NagiosPluginParams]:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        return NagiosPluginParams


    @property
    def params(
        self,
    ) -> NagiosPluginParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = super().params

        assert isinstance(
            params,
            NagiosPluginParams)

        return params


    @property
    def current(
        self,
    ) -> NagiosCurrent:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__current


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.

        :param thread: Child class instance for Chatting Robie.
        """

        assert self.thread

        thread = self.thread
        mqueue = thread.mqueue
        params = self.params

        command = params.command

        names = (
            self.params.clients)


        match: Optional[str]

        kinds = ['privmsg', 'chanmsg']


        if not self.__started:
            self.__started = True
            self.__status('normal')


        while not mqueue.empty:

            mitem = mqueue.get()

            name = mitem.client
            time = mitem.time
            kind = mitem.kind
            family = mitem.family
            isme = mitem.isme
            message = mitem.message

            # Ignore unrelated events
            if kind not in kinds:
                continue

            # Ignore unrelated clients
            if name not in names:
                continue  # NOCVR

            # Ignore event from client
            if isme is True:
                continue

            # Ignore expired messages
            if time.since > 15:
                continue  # NOCVR

            # Basic trust enforcement
            if self.notrust(mitem):
                continue


            match = None

            if family == 'discord':
                match = command.dsc

            if family == 'irc':
                match = command.irc

            if family == 'mattermost':
                match = command.mtm

            # Ignore unrelated clients
            if match is None:
                continue  # NOCVR

            # Ignore unrelated message
            if message != match:
                continue


            if family == 'discord':
                composedsc(self, mitem)

            if family == 'irc':
                composeirc(self, mitem)

            if family == 'mattermost':
                composemtm(self, mitem)


    def trusted(
        self,
        check: Union[str, RobiePerson, RobieMessage],
    ) -> bool:
        """
        Return the boolean indicating whether person is trusted.

        :param check: Validate the person is trusted by plugin.
        :returns: Boolean indicating whether person is trusted.
        """

        params = self.params
        trusted = params.trusted

        if trusted is None:
            return True

        if isinstance(check, RobieMessage):

            if not check.person:
                return False

            check = check.person

        elif isinstance(check, RobiePerson):
            check = check.name

        return check in trusted


    def notrust(
        self,
        check: Union[str, RobiePerson, RobieMessage],
    ) -> bool:
        """
        Return the boolean indicating whether person is trusted.

        :param check: Validate the person is trusted by plugin.
        :returns: Boolean indicating whether person is trusted.
        """

        return not self.trusted(check)


    def __status(
        self,
        status: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param status: One of several possible value for status.
        """

        thread = self.thread
        params = self.params

        if thread is None:
            return None

        plugins = (
            thread.service
            .plugins.childs)

        if 'status' not in plugins:
            return NCNone

        plugin = plugins['status']

        assert isinstance(
            plugin, StatusPlugin)

        (plugin.update(
            unique=self.name,
            group='Extensions',
            title='Nagios Monitor',
            icon=params.status,
            state=status))


    def ainswer(
        self,
    ) -> list[AinswerTool]:
        """
        Return the Ainswer tools that are related to the plugin.

        :returns: Ainswer tools that are related to the plugin.
        """

        from .ainswer import nagios_current

        return [nagios_current]



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.parse import Jinja2
from encommon.parse.jinja2 import FILTERS
from encommon.types import DictStrAny

if TYPE_CHECKING:
    from ..robie import Robie



class RobieJinja2(Jinja2):
    """
    Parse the provided input and intelligently return value.

    :param robie: Primary class instance for Chatting Robie.
    """


    def __init__(
        self,
        robie: 'Robie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        statics: DictStrAny = {
            'robie': robie}

        filters: FILTERS = {}

        super().__init__(
            statics=statics,
            filters=filters)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import get_args

from encommon.types import NCNone

from .ainswer import AinswerIgnored
from .common import AinswerResponseDSC
from .common import AinswerResponseIRC
from .common import AinswerResponseMTM
from .history import AinswerHistoryKinds

if TYPE_CHECKING:
    from .plugin import AinswerPlugin
    from ...robie.models import RobieMessage



_KINDS = get_args(AinswerHistoryKinds)



def composedsc(  # noqa: CFQ004
    plugin: 'AinswerPlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    """

    from ...clients import DSCClient
    from ...clients.discord.message import DSCMessage


    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    childs = robie.childs
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue

    kind = mitem.kind
    hasme = mitem.hasme
    message = mitem.message


    if kind not in _KINDS:
        return NCNone

    if (kind == 'chanmsg'
            and not hasme):
        return None


    assert message is not None

    firschar = (
        message[0].strip())

    if firschar in '!%&/.':
        return NCNone


    assert isinstance(
        mitem, DSCMessage)

    type = mitem.event.type

    if type == 'MESSAGE_UPDATE':
        return NCNone


    client = (
        childs.clients
        [mitem.client])

    assert isinstance(
        client, DSCClient)


    prompt = (
        params.prompt
        .client.dsc)

    ainswer = (
        plugin.ainswer(
            mitem, prompt,
            AinswerResponseDSC))


    if ainswer == AinswerIgnored:
        return NCNone


    citem = mitem.reply(
        robie, ainswer)

    cqueue.put(citem)



def composeirc(  # noqa: CFQ004
    plugin: 'AinswerPlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    """

    from ...clients import IRCClient


    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    childs = robie.childs
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue

    kind = mitem.kind
    hasme = mitem.hasme
    message = mitem.message


    if kind not in _KINDS:
        return NCNone

    if (kind == 'chanmsg'
            and not hasme):
        return None


    assert message is not None

    firschar = (
        message[0].strip())

    if firschar in '!%&/.':
        return NCNone


    client = (
        childs.clients
        [mitem.client])

    assert isinstance(
        client, IRCClient)


    prompt = (
        params.prompt
        .client.irc)

    ainswer = (
        plugin.ainswer(
            mitem, prompt,
            AinswerResponseIRC))


    if ainswer == AinswerIgnored:
        return NCNone


    citem = mitem.reply(
        robie, ainswer)

    cqueue.put(citem)



def composemtm(  # noqa: CFQ004
    plugin: 'AinswerPlugin',
    mitem: 'RobieMessage',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param mitem: Item containing information for operation.
    """

    from ...clients import MTMClient
    from ...clients.mattermost.message import MTMMessage


    assert plugin.thread

    thread = plugin.thread
    robie = plugin.robie
    childs = robie.childs
    params = plugin.params
    member = thread.member
    cqueue = member.cqueue

    kind = mitem.kind
    hasme = mitem.hasme
    message = mitem.message


    if kind not in _KINDS:
        return NCNone

    if (kind == 'chanmsg'
            and not hasme):
        return None


    assert message is not None

    firschar = (
        message[0].strip())

    if firschar in '!%&/.':
        return NCNone


    assert isinstance(
        mitem, MTMMessage)

    type = mitem.event.type

    if type == 'post_edited':
        return NCNone


    client = (
        childs.clients
        [mitem.client])

    assert isinstance(
        client, MTMClient)


    prompt = (
        params.prompt
        .client.mtm)

    ainswer = (
        plugin.ainswer(
            mitem, prompt,
            AinswerResponseMTM))


    if ainswer == AinswerIgnored:
        return NCNone


    citem = mitem.reply(
        robie, ainswer)

    cqueue.put(citem)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from json import dumps
from typing import Any
from typing import Callable
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.times import Time
from encommon.types import DictStrAny
from encommon.types.strings import COMMAS
from encommon.types.strings import NEWLINE
from encommon.types.strings import SEMPTY

from .common import AinswerResponse

if TYPE_CHECKING:
    from .plugin import AinswerPlugin
    from ...robie.childs import RobieClient
    from ...robie.childs import RobiePerson
    from ...robie.models import RobieMessage



AinswerIgnored = 'no_response'

AinswerTool = Callable[..., Any]



@dataclass
class AinswerDepends:
    """
    Dependencies related to operation with PydanticAI tools.
    """

    plugin: 'AinswerPlugin'
    mitem: Optional['RobieMessage']



class AinswerToolset:
    """
    Enumerate the plugins and return those that are related.

    :param plugin: Plugin class instance for Chatting Robie.
    """

    __plugin: 'AinswerPlugin'


    def __init__(
        self,
        plugin: 'AinswerPlugin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__plugin = plugin


    @property
    def toolset(
        self,
    ) -> list[AinswerTool]:
        """
        Return the related tools that were found in the plugins.

        :returns: Related tools that were found in the plugins.
        """

        from .plugin import AinswerPlugin

        plugin = self.__plugin

        toolset: list[AinswerTool] = []

        if not plugin.thread:
            return toolset

        assert plugin.thread

        thread = plugin.thread
        params = plugin.params

        names = params.plugins

        if names is None:
            return toolset


        plugins = (
            thread.service
            .plugins.childs
            .values())

        for helper in plugins:

            name = helper.name

            if name not in names:
                continue

            ignored = isinstance(
                helper, AinswerPlugin)

            if ignored is True:
                continue

            related = hasattr(
                helper, 'ainswer')

            if related is False:
                continue

            assert hasattr(
                helper, 'ainswer')

            toolset.extend(
                helper.ainswer())


        return toolset



class AinswerQuestion:
    """
    Construct prompt and allow for interacting with the LLM.

    :param plugin: Plugin class instance for Chatting Robie.
    """

    __plugin: 'AinswerPlugin'


    def __init__(
        self,
        plugin: 'AinswerPlugin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__plugin = plugin


    def prompt(
        self,
        mitem: 'RobieMessage',
        prompt: str,
    ) -> str:
        """
        Return the message prefixed with runtime prompt values.

        :param mitem: Item containing information for operation.
        :param prompt: Additional prompt insert before question.
        :returns: Message prefixed with runtime prompt values.
        """

        plugin = self.__plugin
        robie = plugin.robie
        params = plugin.params

        _prompt = params.prompt

        header = _prompt.header
        footer = _prompt.footer

        assert mitem.whome

        person = self.__person(mitem)
        client = self.__client(mitem)
        family = mitem.family
        kind = mitem.kind
        whome = mitem.whome


        parsed = robie.j2parse(
            {'prompt': prompt,
             'header': header,
             'footer': footer},
            {'plugin': plugin,
             'client': client,
             'person': person,
             'family': family,
             'kind': kind,
             'whoami': whome[0],
             'mitem': mitem})

        prompt = parsed['prompt']
        header = parsed['header']
        footer = parsed['footer']


        history = (
            self.prompt_history(
                mitem, prompt))

        ignored = (
            self.prompt_ignored(
                mitem, prompt))

        metadata = (
            self.prompt_metadata(
                mitem, prompt))

        channel = (
            self.prompt_channel(
                mitem, prompt))

        recents = (
            self.prompt_recents(
                mitem, prompt))


        returned = SEMPTY.join([
            '**Instructions**\n',
            (f'{prompt}\n\n'
             if prompt
             else SEMPTY),
            (f'{history}\n\n'
             if history
             else SEMPTY),
            (f'{ignored}\n\n'
             if ignored
             else SEMPTY),
            (f'{metadata}\n\n'
             if metadata
             else SEMPTY),
            (f'{channel}\n\n'
             if channel
             else SEMPTY),
            (f'{recents}\n\n'
             if recents
             else SEMPTY),
            (f'{header}\n\n'
             if header is not None
             and len(header) >= 1
             else SEMPTY),
            ('**Question**\n'
             f'{mitem.message}'),
            (f'\n\n{footer}'
             if footer is not None
             and len(footer) >= 1
             else SEMPTY)])


        return returned.strip()


    def prompt_history(
        self,
        mitem: 'RobieMessage',
        prompt: str,
    ) -> str:
        """
        Return the message prefixed with runtime prompt values.

        :param mitem: Item containing information for operation.
        :param prompt: Additional prompt insert before question.
        :returns: Message prefixed with runtime prompt values.
        """

        plugin = self.__plugin
        history = plugin.history

        items: list[DictStrAny] = []

        records = (
            history.records(mitem))

        for record in records:

            person = record.person
            author = record.author
            message = record.message
            ainswer = record.ainswer

            create = (
                Time(record.create)
                .simple)

            items.extend([
                {'role': 'user',
                 'content': message,
                 'nick': author,
                 'user': person,
                 'time': create},
                {'role': 'assistant',
                 'content': ainswer}])

        if len(items) == 0:
            return SEMPTY

        dumped = [
            dumps(x) for x in items]

        return (
            '**Previous Interactions**\n'
            'You have previously had the following'
            ' direct conversations within scope.\n'
            f'{NEWLINE.join(dumped)}')


    def prompt_ignored(
        self,
        mitem: 'RobieMessage',
        prompt: str,
    ) -> str:
        """
        Return the message prefixed with runtime prompt values.

        :param mitem: Item containing information for operation.
        :param prompt: Additional prompt insert before question.
        :returns: Message prefixed with runtime prompt values.
        """

        plugin = self.__plugin
        params = plugin.params

        _prompt = params.prompt
        ignore = _prompt.ignore

        response = AinswerIgnored
        delim = f'{NEWLINE} - '

        return (
            '**Responding**\n'
            'There are reasons not to respond to the'
            ' user question. If you think you should'
            f' not respond, reply only {response}.\n'
            f'Reasons to reply with only {response}'
            f' include:{delim}{delim.join(ignore)}')


    def prompt_metadata(
        self,
        mitem: 'RobieMessage',
        prompt: str,
    ) -> str:
        """
        Return the message prefixed with runtime prompt values.

        :param mitem: Item containing information for operation.
        :param prompt: Additional prompt insert before question.
        :returns: Message prefixed with runtime prompt values.
        """

        assert mitem.whome
        assert mitem.author

        person = self.__person(mitem)
        time = mitem.time
        family = mitem.family
        kind = mitem.kind
        whome = mitem.whome
        author = mitem.author

        returned = (
            '**Message Metadata**\n'
            f'Client family: {family}\n'
            f'Message kind: {kind}\n'
            f'Message time: {time}\n\n'
            '**User Metadata**\n'
            f'User nickname: {author[0]}\n'
            f'User serverID: {author[1]}\n')


        if person is not None:

            returned += (
                'Robie username: '
                f'{person.name}\n')

            _first = person.first
            _last = person.last
            _about = person.about

            if _first is not None:
                returned += (
                    f'First name: {_first}\n')

            if _last is not None:
                returned += (
                    f'Last name: {_last}\n')

            if _about is not None:
                returned += (
                    f'Description: {_about}\n')

        returned += (
            '\n**Additional Metadata**\n'
            f'Your nickname: {whome[0]}\n'
            f'Your serverID: {whome[1]}\n')

        return returned.strip()


    def prompt_channel(
        self,
        mitem: 'RobieMessage',
        prompt: str,
    ) -> str:
        """
        Return the message prefixed with runtime prompt values.

        :param mitem: Item containing information for operation.
        :param prompt: Additional prompt insert before question.
        :returns: Message prefixed with runtime prompt values.
        """

        assert mitem.anchor

        client = self.__client(mitem)
        kind = mitem.kind
        anchor = mitem.anchor

        if kind != 'chanmsg':
            return SEMPTY

        channel = (
            client.channels
            .select(anchor))

        if channel is None:
            return SEMPTY

        unique = channel.unique
        title = channel.title
        topic = channel.topic
        members = channel.members


        returned = (
            '**Channel Metadata**\n'
            f'Unique: {unique}\n')


        if title and title != anchor:
            returned += f'Title: {title}\n'

        if topic is not None:
            returned += f'Topic: {topic}\n'

        if members is not None:
            users = COMMAS.join(members)
            returned += f'Users: {users}\n'


        return returned.strip()


    def prompt_recents(  # noqa: CFQ004
        self,
        mitem: 'RobieMessage',
        prompt: str,
    ) -> str:
        """
        Return the message prefixed with runtime prompt values.

        :param mitem: Item containing information for operation.
        :param prompt: Additional prompt insert before question.
        :returns: Message prefixed with runtime prompt values.
        """

        from ..logger import LoggerPlugin

        plugin = self.__plugin
        robie = plugin.robie
        childs = robie.childs
        plugins = childs.plugins
        params = plugin.params

        name = params.logger
        limit = params.histories

        kind = mitem.kind

        if kind != 'chanmsg':
            return SEMPTY

        if name not in plugins:
            return SEMPTY  # NOCVR

        _plugin = plugins[name]

        if _plugin.name != name:
            return SEMPTY  # NOCVR

        assert isinstance(
            _plugin, LoggerPlugin)

        records = (
            _plugin.history
            .plaintext(limit, mitem))

        return (
            '**Channel Recents**\n'
            'These are the most recent messages from'
            ' the current channel or direct message.'
            ' When client family is IRC or Mattermost,'
            ' you do not see your messages below.\n'
            f'{NEWLINE.join(records)}')


    def __person(
        self,
        mitem: 'RobieMessage',
    ) -> Optional['RobiePerson']:
        """
        Return the person instance where the message originated.

        :param mitem: Item containing information for operation.
        :returns: Person instance where the message originated.
        """

        plugin = self.__plugin
        robie = plugin.robie
        childs = robie.childs
        persons = childs.persons

        _person = mitem.person

        if _person is None:
            return None

        return persons[_person]


    def __client(
        self,
        mitem: 'RobieMessage',
    ) -> 'RobieClient':
        """
        Return the client instance where the message originated.

        :param mitem: Item containing information for operation.
        :returns: Client instance where the message originated.
        """

        plugin = self.__plugin
        robie = plugin.robie
        childs = robie.childs
        clients = childs.clients

        _client = mitem.client

        return clients[_client]


    def submit(
        self,
        prompt: str,
        respond: Type[AinswerResponse],
        mitem: Optional['RobieMessage'] = None,
    ) -> AinswerResponse:
        """
        Submit the question to the LLM and return the response.

        :param prompt: Question that will be asked of the LLM.
        :param respond: Model to describe the expected response.
        :param mitem: Item containing information for operation.
        :returns: Response adhering to provided specifications.
        """

        plugin = self.__plugin

        agent = plugin.agent
        request = agent.run_sync
        model = AinswerDepends

        depends = model(
            plugin=plugin,
            mitem=mitem)

        runsync = request(
            user_prompt=prompt,
            result_type=respond,
            deps=depends)

        return runsync.data



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING
from typing import Union

from encommon.utils.stdout import ANSIARRAY

if TYPE_CHECKING:
    from .childs import RobieClient
    from .childs import RobiePlugin
    from .threads import RobieThreadItems



RobieKinds = Literal[
    'client',
    'plugin',
    'person']

RobiePrint = Union[
    ANSIARRAY,
    'RobieThreadItems']

RobieOperate = Union[
    'RobieClient',
    'RobiePlugin']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING
from typing import Type
from typing import Union
from typing import get_args

from encommon.times import Time
from encommon.types import NCNone

from .common import StatusPluginItem
from .common import StatusPluginStates
from .helpers import composedsc
from .helpers import composeirc
from .helpers import composemtm
from .helpers import reportdsc
from .helpers import reportirc
from .helpers import reportmtm
from .params import StatusPluginIconParams
from .params import StatusPluginParams
from ...robie.childs import RobiePerson
from ...robie.childs import RobiePlugin
from ...robie.models import RobieMessage

if TYPE_CHECKING:
    from .common import StatusPluginItems



_STATES = list(
    get_args(StatusPluginStates))



class StatusPlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin responds to inquiries about Robie status.
    """

    __status: dict[str, StatusPluginItem]
    __stated: dict[str, dict[str, str]]


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__status = {}
        self.__stated = {}


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Nothing to do for plugin


    @classmethod
    def schema(
        cls,
    ) -> Type[StatusPluginParams]:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        return StatusPluginParams


    @property
    def params(
        self,
    ) -> StatusPluginParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = super().params

        assert isinstance(
            params,
            StatusPluginParams)

        return params


    @property
    def status(
        self,
    ) -> 'StatusPluginItems':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__status)


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        assert self.thread

        thread = self.thread
        mqueue = thread.mqueue
        params = self.params

        command = params.command


        match: Optional[str]

        kinds = ['privmsg', 'chanmsg']


        self.report()


        while not mqueue.empty:

            mitem = mqueue.get()

            time = mitem.time
            kind = mitem.kind
            family = mitem.family
            isme = mitem.isme
            message = mitem.message

            # Ignore unrelated events
            if kind not in kinds:
                continue

            # Ignore event from client
            if isme is True:
                continue

            # Ignore expired messages
            if time.since > 15:
                continue  # NOCVR

            # Basic trust enforcement
            if self.notrust(mitem):
                continue


            match = None

            if family == 'discord':
                match = command.dsc

            if family == 'irc':
                match = command.irc

            if family == 'mattermost':
                match = command.mtm

            # Ignore unrelated clients
            if match is None:
                continue  # NOCVR

            # Ignore unrelated message
            if message != match:
                continue


            if family == 'discord':
                composedsc(self, mitem)

            if family == 'irc':
                composeirc(self, mitem)

            if family == 'mattermost':
                composemtm(self, mitem)


    def report(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.
        """

        assert self.thread

        thread = self.thread
        status = self.__status
        stated = self.__stated
        params = self.params

        clients = (
            thread.service
            .clients.childs)

        if not params.reports:
            return NCNone


        def _send_report() -> None:

            if family == 'discord':
                reportdsc(
                    self, client,
                    _status, report)

            if family == 'irc':
                reportirc(
                    self, client,
                    _status, report)

            if family == 'mattermost':
                reportmtm(
                    self, client,
                    _status, report)


        reports = params.reports


        for report in reports:

            name = report.client
            target = report.target
            delay = report.delay
            states = (
                report.states
                or _STATES)

            unique = (
                f'{name}/{target}')

            # Ignore disabled clients
            if name not in clients:
                continue  # NOCVR

            client = clients[name]

            family = client.family

            if unique not in stated:
                stated[unique] = {}

            _stated = stated[unique]


            items = status.items()

            for _unique, _status in items:

                time = _status.time
                since = time.since

                if since < delay:
                    continue  # NOCVR

                state = _status.state

                _state = (
                    _stated.get(_unique))

                if state == _state:
                    continue

                _stated[_unique] = state


                if state not in states:
                    continue  # NOCVR

                _send_report()


    def update(
        self,
        unique: str,
        group: str,
        title: str,
        icon: StatusPluginIconParams,
        state: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param unique: Unique identifier to use for the status.
        :param group: Name for the group the status is membered.
        :param title: Friendly name of the related unique entry.
        :param icon: Optional icon object if supported platform.
        :param state: One of several possible value for status.
        """

        robie = self.robie
        status = self.__status

        object = StatusPluginItem(
            unique=unique,
            time=Time(),
            group=group,
            title=title,
            icon=icon,
            state=state)

        if unique in status:

            _object = status[unique]
            _state = _object.state

            if state == _state:
                return NCNone

        robie.logger.log_i(
            base=self,
            name=self,
            item='status',
            unique=unique,
            state=state)

        status[unique] = object


    def trusted(
        self,
        check: Union[str, RobiePerson, RobieMessage],
    ) -> bool:
        """
        Return the boolean indicating whether person is trusted.

        :param check: Validate the person is trusted by plugin.
        :returns: Boolean indicating whether person is trusted.
        """

        params = self.params
        trusted = params.trusted

        if trusted is None:
            return True

        if isinstance(check, RobieMessage):

            if not check.person:
                return False

            check = check.person

        elif isinstance(check, RobiePerson):
            check = check.name

        return check in trusted


    def notrust(
        self,
        check: Union[str, RobiePerson, RobieMessage],
    ) -> bool:
        """
        Return the boolean indicating whether person is trusted.

        :param check: Validate the person is trusted by plugin.
        :returns: Boolean indicating whether person is trusted.
        """

        return not self.trusted(check)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Literal
from typing import Optional

from pydantic import Field

from ..status import StatusPluginIconParams
from ...robie.params import RobiePluginParams
from ...robie.params.common import RobieParamsModel



_DEFAULT_PROMPT = (
    'You are a helpful chatbot'
    ' assistant named Robie.'
    ' You were built by Robert,'
    ' at the Enasis Network.')

_DEFAULT_IGNORE = [
    ('If you believe that you'
     ' are being abused by the'
     ' user asking the quesiton.')]

_DSC_PROMPT = (
    'In 1875 characters or less,'
    ' answer the user question.'
    ' Format for Discord.'
    ' Markdown is encouraged.')

_IRC_PROMPT = (
    'In 325 characters or less,'
    ' answer the user question.'
    ' Format for IRCv2.'
    ' Do not use markdown.'
    ' Do not use colors.')

_MTM_PROMPT = (
    'In 1875 characters or less,'
    ' answer the user question.'
    ' Format for Mattermost.'
    ' Markdown is encouraged.')



class AinswerPluginAinswerParams(RobieParamsModel, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    origin: Annotated[
        Literal['anthropic', 'openai'],
        Field(...,
              description='Which platform hosts the model')]

    model: Annotated[
        str,
        Field(...,
              description='Platform model that will be used',
              min_length=1)]

    secret: Annotated[
        str,
        Field(...,
              description='Model in platform that is used',
              min_length=1)]

    timeout: Annotated[
        int,
        Field(30,
              description='Time to wait during the request',
              ge=1, le=300)]

    sleep: Annotated[
        tuple[int, int],
        Field((15, 30),
              description='Time to wait before the request')]



class AinswerPluginPromptClientParams(RobieParamsModel, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    dsc: Annotated[
        str,
        Field(_DSC_PROMPT,
              description='Supplement the system prompt',
              min_length=1)]

    irc: Annotated[
        str,
        Field(_IRC_PROMPT,
              description='Supplement the system prompt',
              min_length=1)]

    mtm: Annotated[
        str,
        Field(_MTM_PROMPT,
              description='Supplement the system prompt',
              min_length=1)]



class AinswerPluginPromptParams(RobieParamsModel, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    system: Annotated[
        str,
        Field(_DEFAULT_PROMPT,
              description='Override the agent system prompt',
              min_length=1)]

    client: Annotated[
        AinswerPluginPromptClientParams,
        Field(default_factory=AinswerPluginPromptClientParams,
              description='Additional chat platform prompt')]

    header: Annotated[
        Optional[str],
        Field(None,
              description='Optional header before question',
              min_length=1)]

    footer: Annotated[
        Optional[str],
        Field(None,
              description='Optional footer after question',
              min_length=1)]

    ignore: Annotated[
        list[str],
        Field(_DEFAULT_IGNORE,
              description='Reasons for LLM to decline response',
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        if _parse is not None:

            parsable = [
                'system',
                'ignore']

            for key in parsable:

                value = data.get(key)

                if value is None:
                    continue

                value = _parse(value)

                data[key] = value


        super().__init__(**data)



class AinswerPluginParams(RobiePluginParams, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    database: Annotated[
        str,
        Field('sqlite:///:memory:',
              description='Database connection string',
              min_length=1)]

    histories: Annotated[
        int,
        Field(10,
              description='Number of messages per anchor',
              ge=1, le=1000)]

    clients: Annotated[
        list[str],
        Field(...,
              description='List of clients to enable plugin',
              min_length=1)]

    ainswer: Annotated[
        AinswerPluginAinswerParams,
        Field(...,
              description='Parameters for the AI platforms')]

    prompt: Annotated[
        AinswerPluginPromptParams,
        Field(default_factory=AinswerPluginPromptParams,
              description='Override the agent system prompt')]

    plugins: Annotated[
        Optional[list[str]],
        Field(None,
              description='List of plugins to load tools',
              min_length=1)]

    logger: Annotated[
        Optional[str],
        Field(None,
              description='Logger for including recents',
              min_length=1)]

    trusted: Annotated[
        Optional[list[str]],
        Field(None,
              description='Users are trusted by the plugin',
              min_length=1)]

    status: Annotated[
        StatusPluginIconParams,
        Field(default_factory=StatusPluginIconParams,
              description='Icon used per chat platform')]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        clients = data.get('clients')

        if isinstance(clients, str):
            data['clients'] = [clients]


        if _parse is not None:

            parsable = ['prompt']

            for key in parsable:

                if not data.get(key):
                    continue

                item = data[key]

                item['_parse'] = _parse


            parsable = [
                'database',
                'histories',
                'ainswer',
                'logger',
                'status']

            for key in parsable:

                value = data.get(key)

                if value is None:
                    continue

                value = _parse(value)

                data[key] = value


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .command import RobieCommand
from .message import RobieMessage
from .message import RobieMessageKinds
from .robie import RobieModels



__all__ = [
    'RobieModels',
    'RobieMessage',
    'RobieMessageKinds',
    'RobieCommand']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from enconnect.fixtures import DSCClientSocket
from enconnect.fixtures import IRCClientSocket
from enconnect.fixtures import MTMClientSocket

from pydantic_ai.models.test import TestModel

from ..plugin import AinswerPlugin
from ....clients.discord.test import DSCEVENTS
from ....clients.irc.test import IRCEVENTS
from ....clients.mattermost.test import MTMEVENTS

if TYPE_CHECKING:
    from ....robie import RobieService



def test_AinswerPlugin(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    plugin = (
        service.plugins
        .childs['ainswer'])

    assert isinstance(
        plugin, AinswerPlugin)


    attrs = lattrs(plugin)

    assert attrs == [
        '_RobieChild__robie',
        '_RobieChild__name',
        '_RobieChild__params',
        '_AinswerPlugin__started',
        '_AinswerPlugin__toolset',
        '_AinswerPlugin__question',
        '_AinswerPlugin__history',
        '_AinswerPlugin__model',
        '_AinswerPlugin__agent',
        '_RobiePlugin__thread']


    assert inrepr(
        'plugin.AinswerPlugin',
        plugin)

    assert isinstance(
        hash(plugin), int)

    assert instr(
        'plugin.AinswerPlugin',
        plugin)


    plugin.validate()

    assert plugin.robie

    assert plugin.enable

    assert plugin.name == 'ainswer'

    assert plugin.kind == 'plugin'

    assert plugin.schema()

    assert plugin.params

    assert plugin.thread

    assert plugin.dumped

    assert plugin.toolset

    assert plugin.question

    assert plugin.history



def test_AinswerPlugin_cover(
    service: 'RobieService',
    client_dscsock: DSCClientSocket,
    client_ircsock: IRCClientSocket,
    client_mtmsock: MTMClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    :param client_dscsock: Object to mock client connection.
    :param client_ircsock: Object to mock client connection.
    :param client_mtmsock: Object to mock client connection.
    """

    robie = service.robie
    childs = robie.childs
    clients = childs.clients

    plugin = (
        service.plugins
        .childs['ainswer'])

    assert isinstance(
        plugin, AinswerPlugin)

    history = plugin.history


    client_dscsock(DSCEVENTS)
    client_ircsock(IRCEVENTS)
    client_mtmsock(MTMEVENTS)


    testing = TestModel()

    override_agent = (
        plugin.agent
        .override(
            model=testing))

    with override_agent:


        service.limit(
            plugins=[
                'ainswer',
                'logger',
                'status'])

        service.start()


        thread = Thread(
            target=service.operate)

        thread.start()


        block_sleep(5)


        select = (
            clients['ircbot']
            .channels
            .select('#enrobie'))

        assert select is not None

        records = (
            history.search(
                client='ircbot',
                anchor='#enrobie'))

        assert len(records) >= 1


        service.soft()

        while service.running:
            block_sleep(1)

        service.stop()

        thread.join()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .ainswer import AinswerDepends
from .ainswer import AinswerTool
from .params import AinswerPluginParams
from .plugin import AinswerPlugin



__all__ = [
    'AinswerPlugin',
    'AinswerPluginParams',
    'AinswerDepends',
    'AinswerTool']
